package HTTP::Server::Zaphod;
use strict;
use base 'HTTP::Server::Sajax';
use Zaphod::Session;
use Zaphod::Session::Multiplexer;
use File::Glob qw(bsd_glob);

use vars qw($VERSION);
$VERSION = '0.02';

__PACKAGE__->mk_accessors(qw(
    session_dir
    request_count
    save_count
));

=head1 NAME

HTTP::Server::Zaphod - HTTP server for Zaphod

=head1 SYNOPSIS

#!/usr/bin/perl -w
use lib 'lib';
use HTTP::Server::Zaphod;
use Zaphod::Session;

    my $httpd = HTTP::Server::Zaphod->new(
        template    => 'codeview.tt',
        static      => 'static',
        port        => 8080,
        document    => "Hello World\nThis is a test\nwith some lines\n",
        session_dir => 'sessions',
        save_count  => 500,
    );
    #close STDERR;
    #$httpd->background();
    $httpd->run();

=cut

=head1 METHODS

=head2 C<< HTTP::Server::Zaphod->new ARGS >>

Constructs a new server from the arguments given.

The following keys are valid:

  document      - the document every L<Zaphod::Editor>
                  instance will be initialized with
  session_dir   - the directory where sessions will be
                  serialized into
  save_count    - the number of requests after which
                  all sessions are auto-saved

The remaining arguments will be passed onto L<HTTP::Server::Sajax>.

=cut

sub new {
    my ($package, %args) = @_;

    my $s = Zaphod::Session::Multiplexer->new(
        session_args => [
            document => delete $args{ document },
        ],
    );
    
    $args{request_count} = 0
        unless exists $args{request_count};
    if (exists $args{session_dir}) {
        # Auto-save every 500 handled requests
        $args{save_count} ||= 500;
        warn "Saving every $args{save_count}";

        $s->load_session( $_ ) 
            for glob "$args{session_dir}/*.meta";

    } else {
        # Create one fresh session
        $s->new_session();
    };
    
    my $self = __PACKAGE__->SUPER::new(
        object   => $s,
        methods  => [qw[ set_name users document connect_client disconnect_client actions]],
        static => 'static',
        %args,
    );
    $self->save_count($args{save_count});
    $self->session_dir($args{session_dir});
    $self;
};

sub handle_request {
    my ($self,$q) = @_;
    
    # Check if we want to auto-save
    my $c = $self->request_count;
    $self->request_count( $c + 1 );
    warn "Saving all sessions $c/". $self->save_count;
    if (defined $self->save_count and $c % $self->save_count == 0) {
        $self->module->save_sessions( $self->session_dir );
    }
    
    binmode $self->stdout_handle, ":utf8";
    my $path = $q->url( -absolute => 1, -path_info => 1 );
    if ($path =~ m!^/session/new$!) {
        my $s = $self->module->new_session();
        # XXX Make it so it uses a relocatable base
        print "HTTP/1.0 302 Moved\r\n";
        print $q->redirect("/session/" . $s->cookie);
    } elsif ($path =~ m!^/session/(.+)$!) {
        $self->serve_ajax($q, session => $1 );
    } elsif ($path =~ m!^/(?:session)?$!) {
        $self->serve_sessions($q);
    } elsif ($path =~ m!^/static/(t/|jsan(/\w+)*/)?[\w.]+$!) {
        $self->serve_static( $q, $self->static . "/.." );
    } else {
        $self->serve_ajax($q);
    };
};

sub serve_ajax {
    my ($self,$q,%args) = @_;

    my $s = $self->module->sessions->{ $args{ session } };
    my $usercount;
    if ($s) {
        $usercount = $s->user_count;
    };
    $self->SUPER::serve_ajax($q,
        session => $s,
        user_count => $usercount,
        request => $q,
    );
};

sub serve_sessions {
    my ($self,$q) = @_;
    $self->_template->process($self->session_template,
    {
            request   => $q,
            module    => $self->module,
    });
    warn "TEMPLATE: " . $self->_template->error
        if $self->_template->error;
};

1;

=head1 AUTHOR

Written by Max Maischein E<lt>corion@cpan.orgE<gt>

Copyright Max Maischein 2005,2006; All Rights Reserved.

This program is free software; you can redistribute it
and/or modify it under the same terms as Perl itself.

=head1 SEE ALSO

L<Zaphod>

=cut
