package Zaphod::Editor::Action::InsertLine;
use strict;
use base qw'Zaphod::Editor::Action';

use Data::Dumper;
use Carp qw(croak);

use vars qw($VERSION);
$VERSION = '0.02';

__PACKAGE__->verbs([qw(
    insert_line
    delete_line
    replace_line
    split_line
    join_line
)]);


sub insert_line {
    my ($class,$editor,@lines) = @_;

    my $status = $editor->status;
    my $status_after = $status->move_down( scalar @lines );
    my $nl = $editor->newline;
    my $text = join( $nl, @lines) . $nl;

    $class->new(
        editor => $editor,
        status_before => $status,
        status_after => $status_after,
        text_before => '',
        text_after => $text,
    )
};

sub delete_line {
    my ($class,$editor,$count) = @_;

    my $status = $editor->status;
    my $status_after = $status->copy;
    die __PACKAGE__ . "::delete_line is unimplemented.";

    my $old_text = '<restore unimplemented>';

    $class->new( 
        editor => $editor,
        status_before => $status,
        status_after => $status_after,
        text_before => $old_text,
        text_after => '')
};

sub replace_line {
  my ($class,$editor,$count,@lines) = @_;
  my $status = $editor->status;
  my $i = $status->copy;

  $count = 1 unless defined $count;
  my $status_after = $status->move_down(scalar @lines);

  # Needs to know about joining/parting lines
  my $lines = [ @{$editor->document}[ $status->line..$status->line+$count-1 ]];

  $class->new( editor => $editor, insertion_point => $i,
               status_before => $status, status_after => $status_after, text_before => $lines, text_after => \@lines)
};

sub split_line {
    my ($class,$editor) = @_;

    # Somehow, an auto-indent hook should be provided here,
    # or rather, somewhere else. Look at how vi / emacs do this.

    Zaphod::Editor::Action::InsertChar->new($editor, $editor->newline);
}

sub join_line {
    my ($class,$editor) = @_;
    my $joined = $editor->current_line;
    my $sep = "";
    my $s = $editor->status;

    # Needs document bounds check
    # Should be more configurable, as in, the separator should be configurable
    my $right = $editor->document->[ $editor->status->line +1 ];
    if ($joined !~ /\s$/ and $right ne "" and $right !~ /^\s/) {
        $sep = " ";
    }
    $joined .= $sep . $right;
    my $a = $class->replace_line($editor,2, $joined);

    $a->status_before($s->copy);
    $a->status_after($s->copy);
    $a->insertion_point($s->copy);

    $a
}

sub as_string {
    my ($self) = @_;
    join "\n", $self->insertion_point->as_string,
               (map { "< $_" } @{ $self->text_before }),
               (map { "> $_" } @{ $self->text_after }),
};

sub as_linemap {
    my ($self) = @_;
    my $ipl = $self->insertion_point->line;
    sprintf "$ipl:(%s) => $ipl:(%s)", join(",", $self->affected_lines_pre), join(",", $self->affected_lines_post);
};

=head2 C<< $action->get_line_pre LINENUMBERS >>

Maps the line numbers after the action to those I<before>
the action was applied.

=cut

sub get_line_pre {
    my ($self,@numbers) = @_;

    my $first = $self->insertion_point->line;
    my $last  = $first + @{ $self->text_after };

    my $diff  = @{ $self->text_after } - @{ $self->text_before };

    wantarray
    ?   map {
              $_ >= $first && $_ < $last ? undef
            : $_ >= $last                ? $_ - $diff
            : $_
        } (@numbers)
    :         $numbers[0] >= $first && $numbers[0] < $last ? undef
            : $numbers[0] >= $last                         ? $numbers[0] - $diff
            : $numbers[0]
};

=head2 C<< $action->get_line_post LINENUMBERS >>

Maps the line numbers before the action to those I<after>
the action was applied.

=cut

sub get_line_post {
    my ($self,@numbers) = @_;

    my $first = $self->insertion_point->line;
    my $last  = $first + @{ $self->text_before };
    my $diff  = @{ $self->text_after } - @{ $self->text_before };

    wantarray
    ?   map {
            ! defined($_)                ? undef
            : $_ >= $first && $_ < $last ? undef
            : $_ >= $last                ? $_ + $diff
            : $_
        } (@numbers)
    :         $numbers[0] >= $first && $numbers[0] < $last ? undef
            : $numbers[0] >= $last                         ? $numbers[0] + $diff
            : $numbers[0]
};

=head2 C<< $action->commutes_with ACTION >>

This is a specialized implementation implementing the following
commutation rules, for line-based actions only!

The following actions commute:

  delete + delete
  insert + nonoverlapping insert
  
  insert + disjunct delete

The following actions do not commute:

  insert + internal insert
  insert + overlapping delete

=cut

sub commutes_with {
    my ($left,$right) = @_;
    if (($left->start_line < $right->start_line and $left->end_line < $right->start_line)
     or ($right->start_line < $left->start_line and $right->end_line < $left->start_line)) {
        # The two actions share no overlap, so they commute
        return 1
    } else {
         return
    }
}

=head1 TODO

=over 4

=item *

Should be able to return list of actions instead of
just one action.

=item *

Implement C<insert_line> and C<delete_line> in terms of
C<replace_line>

=item *

Implement joining of lines, for C<delete_left> / C<delete_right>,
and maybe as a separate command too.

=item *

Implement insert-lines-in-the-middle instead of only above
or below the current line

=back

=cut

1;

=head1 AUTHOR

Written by Max Maischein E<lt>corion@cpan.orgE<gt>

Copyright Max Maischein 2005,2006; All Rights Reserved.

This program is free software; you can redistribute it
and/or modify it under the same terms as Perl itself.

=head1 SEE ALSO

L<Zaphod>

=cut
