package Zaphod::Session::API::JS;
use strict;
use base 'Class::Accessor';
use Data::Dumper;
use Encode qw(decode);
use Zaphod::Session::Response;

use vars qw($VERSION);
$VERSION = '0.02';

__PACKAGE__->mk_accessors(qw(
    session
));

=head2 C<< $session->connect_client VISUAL, STATE >>

This is the JS accessible version of the login which
returns a JS string to set up the connection.

Also sends the diffs from STATE to the current state
if C<STATE> is 0 or C<undefined>. If C<STATE> is -1
then no updates are sent as the client is expected
to retrieve the complete document via C<document>.

=cut

sub connect_client {
    my ($self,$visual,$state) = @_;
    $state ||= 0;
    $visual = $self->params2utf8( $visual );
    my $s = $self->session;
    my $client = $s->connect_client($visual);
    
    my @changes;
    if ($state >= 0) {
        @changes = $s->actions_for_client($client,$state);
    }
    my $r = Zaphod::Session::Response->new({ 
        state => $s->state,
        changes => \@changes,
        user => $client,
        delay => 1000, # ms
        cookie => $client,
    });
    
    return $self->js_response($r);
}

=head2 C<< $api->valid_cookie COOKIE >>

(Internal only)

Returns the user if the cookie is a valid
cookie for this session. See also
L<Zaphod::Session>::valid_cookie

=cut

sub valid_cookie {
    $_[0]->session->valid_cookie($_[1]);
}

=head2 C<< $session->new_user CLIENT >>

Returns C<CLIENT> as a C<new User()> JS statement,
used by C<connect_client>.

=cut

sub new_user {
    my ($self,$client) = @_;
    $client = $self->session->clients->{$client};
    die "No client found ($client)"
        unless $client;
    return 
        sprintf "new User(%s,'%s',%s)",
            $client->{color},
            $client->{visual},
            $client->{cursor_position},
        ;
}

=head2 C<< $session->disconnect_client CLIENT >>

Removes the client from the session

=cut

sub disconnect_client {
    my ($self,$client) = @_;
    $self->session->disconnect_client($client);
    my $r = Zaphod::Session::Response->new({});
    return $self->js_response($r);
}

=head2 C<< $session->users CLIENT, OFS, STATE >>

Returns the (positions of the) other users
in the current session as Javascript coded.

Also returns any new actions that happened since
C<STATE>, if any.

TODO: All of C<changes>, C<users>, C<document> and whatnot
should be rolled into two structures, one input
and one output structure.

=cut

sub users {
    my ($self,$client,$ofs,$state) = @_;
    my $s = $self->session;
    
    if (! $s->valid_cookie($client)) {
        my $r = Zaphod::Session::Response->new({
            code => "do_logout()",
        });
        return $self->js_response($r);
    }
    
    $s->client_cursor($client,$ofs);
    my @users = $s->users_for_client($client);
    
    my @state;
    my @changes;
    if ($state eq "undefined" or $state != $s->state) {
        $state = 0 + $state;
        @changes = $s->actions_for_client($client,$state);
        @state = ("state",$s->state);
    }
    
    my @session_name;
    if ($state <= $s->name->[0]) {
        # Send the new name. Maybe we should send the changing user too.
        #my $js = sprintf qq{setSessionName(%s)}, 
        #                          $self->encode_js_string($s->name->[1]);
        @session_name = ("session",$s->name->[1]);
    };
    my $r = Zaphod::Session::Response->new({
        users => \@users,
        changes => \@changes,
        @state,
        @session_name,
    });
    
    return $self->js_response($r);
}

=head2 C<< $session->document >>

Returns a JS statement to assign the document.

TODO: The "current" document should be a C<text/plain>
document that can be retrieved partially to speed up
the handling/editing of large documents.

=cut

sub document { 
    my $r = Zaphod::Session::Response->new({
        document => $_[0]->session->editor->document,
    });
    $_[0]->js_response($r);
}

=head2 C<< $session->set_name CLIENT, STATE, NEWNAME >>

Sets the name of the current session, permissions provided.

The new name will be broadcast to the other clients.
The name change is currently out-of-band with the editor,
so there is no undo/redo functionality for it.

=cut

sub set_name {
    my ($self,$client,$state,$name) = @_;
    $self->session->set_name($client, $state, $name);
    my $r = Zaphod::Session::Response->new({ session => $name });
    $self->js_response($r);
}

=head2 C<< $session->js_response RESPONSE, @code >>

Returns the statements wrapped as a JS response.

Always prepends a timestamp to the C<timestamp>
JS variable.

TODO: The client could check/maintain the time of the
last response itself.

=cut

sub js_response {
    my ($self,$response) = @_;
    my @items;
    $response->delay(1000);
    push @items, '"timestamp":' . $response->timestamp;
    push @items, '"state":' . $response->state
        if defined $response->state;
    push @items, '"user":' . $self->new_user( $response->user )
        if ($response->user);
    push @items, '"delay":' . $response->delay
        if ($response->delay);
    push @items, '"code":' . join "", $self->encode_js_string($response->code)
        if ($response->code);
    push @items, '"cookie":' . join "", $self->encode_js_string($response->cookie)
        if ($response->cookie);
    push @items, '"document":' . join "", $self->encode_js_string($response->document)
        if (defined $response->document);

    if ($response->session) {
        my $s = $self->encode_js_string($response->session);
        push @items, '"session":' . $s;
    }

    if ($response->users and my @u = @{ $response->users }) {
        push @items, '"users":[' . join(",", map { $self->new_user($_)} @u) . "]";
    };

    if ($response->changes and my @c = @{$response->changes}) {
        push @items, '"changes":[' . join( ",", map { $_->as_js_string } @c) . "]";
    };
    
    my $js = "{\n" . join( ",\n", @items ). "\n}";
    warn "js_response(): $js";
    return $js;
}

=head2 C<< $sesion->actions CLIENT, STATE, ACTIONSTRING >>

Takes a string of actions, reconciles them with the
parallel other actions, and returns JS code to update
the client positions and to replace the submitted
actions. Also returns code to update the client
state number.

TODO: Currently knows too much about the underlying
C<session> object. Separate the functionalities.

=cut

sub actions {
    my ($self,$client,$ofs,$state,$action_str) = @_;
    warn "$client / $ofs / $state / $action_str";
    $action_str = $self->params2utf8( $action_str );
    
    if ($state eq 'undefined') {
        $state = 0;
    };
    
    my $e = $self->session->editor;
    my @changes = Zaphod::Editor::Action->from_string($e,$action_str);
    
    my $old_count = @changes;
    my $visual = $self->session->clients->{$client}->{visual};
    warn @changes . " actions from $visual ($action_str)";

    my $new_state = $self->session->action_from_client($client,$state,@changes);
    my @diff_actions = $self->session->actions_for_client($client,$state);
    my @users = $self->session->users_for_client($client);

    warn sprintf "Sending back %s actions to %s", scalar @diff_actions, $visual;
    my $r = Zaphod::Session::Response->new({
        users => \@users,
        changes => \@diff_actions,
        state => $new_state,
    });
    return $self->js_response( $r );
}

=head2 C<< $session->encode_js_string STRINGS >>

Helper method to generate quoted JS strings. The
nonprintable or special characters will be escaped
by using backslashes and hexadecimal notation (C<\xFF>).

TODO: Use a standard (JSON?) method instead of hand-rolling my own.

=cut

sub encode_js_string {
    my $self = shift;
    wantarray ?
    map {
        s/([\x00-\x1f\\"\x80-\xFF])/sprintf "\\x%02x", ord $1/gme;
        qq{"$_"}
    } @_
    : do {
        my $x = $_[0];
        $x =~ s/([\x00-\x1f\\"\x80-\xFF])/sprintf "\\x%02x", ord $1/gme;
        qq{"$x"}
    };
}

=head2 C<< $session->params2utf8 ARGS >>

Converts all the arguments from ISO-8859-1 to utf8.

=cut

sub params2utf8 {
    my $self = shift;
    my $encoding = "ISO-8859-1";
    wantarray
    ? map { decode($encoding, $_   )} @_
    :       decode($encoding, $_[0])
}

1;