package Zaphod::Session::Multiplexer;
use strict;
use base 'Class::Accessor';
use Digest::MD5 qw( md5_base64 );
use URI::Escape;
use Carp qw(carp);
use Data::Dumper;

use vars qw($VERSION);
$VERSION = '0.02';

__PACKAGE__->mk_accessors(qw(
    sessions
    session_key_seed
    session_args
));

=head1 NAME

Zaphod::Session::Multiplexer - Multiplexer for JS requests

=cut

sub new {
    my ($class,%args) = @_;
    $args{sessions} ||= {};
    $args{session_key_seed} ||= "Another Zaphod Secret";
    $args{session_args} ||= []; # Use defaults of Zaphod::Session
    $class->SUPER::new(\%args);
}

=head2 C<< $session->gen_session_key SESSION_ARGS >>

Generates a string for uniquely/opaquely identifying
a session.

=cut

sub gen_session_key {
    my ($self,$args) = @_;
    
    # This should become configurable, later,
    # or simply initialized at server startup
    # to some random string
    my $secret = $self->session_key_seed;

    return uri_escape(md5_base64( "$secret$$" . rand(0x10000)));
}

=head2 C<< $mx->add_session KEY, ARGS >>

Adds new session to the list of sessions.

=cut

sub add_session {
    my ($self,$s) = @_;
    my $key = $s->cookie;
    $self->sessions->{$key} = $s;
    warn "Adding a session, now at " . scalar keys %{ $self->sessions };
    $s
}

=head2 C<< $mx->save_sessions DIR >>

Saves all sessions into the given directory.

=cut

sub save_sessions {
    my ($self,$dir) = @_;
    warn "Saving all sessions";
    for (values %{ $self->sessions }) {
        $_->save($dir);
    }
}

=head2 C<< $mx->load_session FILE >>

Loads a session from a C<.meta> file
and adds it to the list. Returns
the newly created session.

=cut

sub load_session {
    my ($self,$fn) = @_;
    my $s = Zaphod::Session->load( $fn, @{ $self->session_args });
    $self->add_session( $s );
}

=head2 C<< $mx->session_for_client >>

Returns the session a client belongs to.

=cut

sub session_for_client {
    my ($self,$cookie) = @_;
    my $all = $self->sessions;
    for my $s (sort values %$all) {
        if ($s->valid_cookie($cookie)) {
            return $s
        }
    }
}

=head2 C<< $mx->new_session DOCUMENT_ID >>

Returns a new session as initialized
from the C<session_args>. The session
is automatically added to the list
of active sessions.

If C<DOCUMENT_ID> is a false value,
a document id will be generated
via C<gen_session_key>.

=cut

sub new_session {
    my ($self,$document_id) = @_;
    $document_id ||= $self->gen_session_key;
    my $s = Zaphod::Session->new(cookie => $document_id, @{ $self->session_args });
    $self->add_session($s);
}

=head2 document VISUAL, SESSION_KEY, STATE

Forwards the request for the document to
the corresponding session.

If no fitting session is found,
redirects to the list of sessions. Well,
maybe later.

XXX

=cut

sub document {
    my ($self,$document_id,@args) = @_;
    $document_id ||= $self->gen_session_key;
    my $s = $self->sessions->{$document_id};
    warn $s;
    if (! $s) {
        my $r = Zaphod::Session::Response->new({
            message => "Session does not exist",
            code    => "document.location = '/session'",
        });
        return Zaphod::Session::API::JS->js_response($r);
    };
    return $s->api('JS')->document(@args);
}

=head2 C<< $mx->connect_client SESSION_KEY, ARGS >>

Connects a client to the specified session. If the
session doesn't exist or is a false value, a new
session is created and the client is connected
to it.

The parameters for the session are taken from
the C<session_args> array.

=cut

sub connect_client {
    my ($self,$document_id,@args) = @_;
    $document_id ||= $self->gen_session_key;
    my $s = $self->sessions->{$document_id};
    if (! $s) {
        $s = $self->new_session($document_id);
    };
    $s->api('JS')->connect_client(@args);
}

=head2 disconnect_client

=head2 users

=head2 actions

If a session for the client cannot be found,
it is rejected, otherwise these methods are 
forwarded to the implementation
in L<Zaphod::Session::API::JS>.

=cut

for (qw( disconnect_client users actions set_name )) {
    my $method = $_;
    my $name = __PACKAGE__ . "::$_";
    no strict 'refs';
    *{ $name } = sub {
        my ($self,$cookie,@args) = @_;
        my $session = $self->session_for_client($cookie);
        if ($session) {
            $session->api('JS')->$method($cookie,@args);
        } else {
            my $r = Zaphod::Session::Response->new({
                code   => "do_logout()",
                reason => "Invalid session",
            });
            return Zaphod::Session::API::JS->js_response($r);
        }
    };
}

=head1 TODO

=over 4

=item *

The name is badly chosen. It should be a multiplexer
on a different level. Ah well.

=item *

Currently a client is always tied to a certain document.

=item *

Handle clients here instead of within Zaphod::Session
because client connections exist across sessions/documents.

=back

=cut

1;